﻿using System;
using System.IO;
using System.Xml.Linq;

namespace compiler
{
    public class WorldWriter : IWriter
    {
        public void Write(string path, TestParser.Node rootNode)
        {
            var world = CreateBaseWorld();

            bool vehicleFound = false;
            foreach (var node in rootNode)
            {
                if (node.Text.StartsWith("spawn:", StringComparison.CurrentCultureIgnoreCase))
                {
                    var text = node.Text.Substring("spawn:".Length).Trim();
                    if (text.StartsWith("Vehicle", StringComparison.CurrentCultureIgnoreCase))
                    {
                        if (vehicleFound)
                        {
                            ConsoleLogger.WriteLine("Multiple Vehicles found", Verbosity.Error);
                            throw new ProcessingException();
                        }
                        vehicleFound = true;//spawn the vehicle always at 0 0 0
                    }
                    else if (text.StartsWith("Obstacle", StringComparison.CurrentCultureIgnoreCase))
                    {
                        var obstacle = this.CreateObstacle(new Obstacle(text));
                        world.Add(obstacle);
                    }
                    else
                    {
                        ConsoleLogger.WriteLine("Unknown spawn-Type: " + text, Verbosity.Error);
                        throw new ProcessingException();
                    }
                }
            }

            if (!vehicleFound)
            {
                ConsoleLogger.WriteLine("No Vehicles found", Verbosity.Error);
                throw new ProcessingException();
            }

            var doc = this.CreateDocumentFromWorld(world);
            var targetFile = Path.Combine(path, "test.world");
            doc.Save(targetFile);
        }

        private XDocument CreateDocumentFromWorld(XElement world)
        {
            var doc = new XDocument(XElement.Parse("<sdf version='1.4'></sdf>"));
            doc.Root.Add(world);
            return doc;
        }

        private XElement CreateBaseWorld()
        {
            return XElement.Parse(@"
<world name='default'>
<!-- A global light source -->
<include>
    <uri>model://sun</uri>
</include>

<!-- A ground plane -->
<include>
    <uri>model://ground_plane</uri>
</include>

<physics type='ode'>
    <real_time_update_rate>1000.0</real_time_update_rate>
    <max_step_size>0.001</max_step_size>
    <real_time_factor>1</real_time_factor>
    <ode>
    <solver>
        <type>quick</type>
        <iters>150</iters>
        <precon_iters>0</precon_iters>
        <sor>1.400000</sor>
        <use_dynamic_moi_rescaling>1</use_dynamic_moi_rescaling>
    </solver>
    <constraints>
        <cfm>0.00001</cfm>
        <erp>0.2</erp>
        <contact_max_correcting_vel>2000.000000</contact_max_correcting_vel>
        <contact_surface_layer>0.01000</contact_surface_layer>
    </constraints>
    </ode>
</physics>

<scene>
    <ambient>0.4 0.4 0.4 1</ambient>
    <background>0.7 0.7 0.7 1</background>
    <shadows>true</shadows>
</scene>

<gui fullscreen='0'>
    <camera name='user_camera'>
    <pose>0.0 0.0 17.0 0 1.5708 0</pose>
    <view_controller>orbit</view_controller>
    </camera>
</gui>
</world>");
        }

        private XElement CreateObstacle(Obstacle obstacle)
        {
            switch (obstacle.Type)
            {
                case "box":
                    return XElement.Parse(@$"<model name='box_{obstacle.Name}_R_{obstacle.Parameter["l"]}'>
<pose frame=''>{obstacle.Parameter["x"]} {obstacle.Parameter["y"]} 0 0 0 0</pose>
<link name='link'>
<inertial>
    <mass>1</mass>
    <inertia>
    <ixx>0.166667</ixx>
    <ixy>0</ixy>
    <ixz>0</ixz>
    <iyy>0.166667</iyy>
    <iyz>0</iyz>
    <izz>0.166667</izz>
    </inertia>
</inertial>
<collision name='collision'>
    <geometry>
    <box>
        <size>{obstacle.Parameter["l"]} {obstacle.Parameter["l"]} {obstacle.Parameter["l"]}</size>
    </box>
    </geometry>
    <max_contacts>10</max_contacts>
    <surface>
    <contact>
        <ode/>
    </contact>
    <bounce/>
    <friction>
        <torsional>
        <ode/>
        </torsional>
        <ode/>
    </friction>
    </surface>
</collision>
<visual name='visual'>
    <geometry>
    <box>
        <size>{obstacle.Parameter["l"]} {obstacle.Parameter["l"]} {obstacle.Parameter["l"]}</size>
    </box>
    </geometry>
    <material>
    <script>
        <name>Gazebo/Grey</name>
        <uri>file://media/materials/scripts/gazebo.material</uri>
    </script>
    </material>
</visual>
<self_collide>0</self_collide>
<enable_wind>0</enable_wind>
<kinematic>0</kinematic>
</link>
</model>");
                case "cylinder":
                    return XElement.Parse(@$"<model name='cylinder_{obstacle.Name}_R_{obstacle.Parameter["r"]}'>
<pose>{obstacle.Parameter["x"]} {obstacle.Parameter["y"]} 0 0 0 0</pose>
<link name='link'>
<collision name='collision'>
    <geometry>
    <cylinder>
        <radius>{obstacle.Parameter["r"]}</radius>
        <length>{obstacle.Parameter["h"]}</length>
    </cylinder>
    </geometry>
</collision>
<visual name='visual'>
    <geometry>
    <cylinder>
        <radius>{obstacle.Parameter["r"]}</radius>
        <length>{obstacle.Parameter["h"]}</length>
    </cylinder>
    </geometry>
</visual>
</link>
</model>");
                //case "sphere":
                //case "ball"
            }

            ConsoleLogger.WriteLine("Unknown Obstacle Type '" + obstacle.Type + "'", Verbosity.Error);
            throw new ProcessingException();
        }
    }
}
